#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

double *X, *Y, *Z;     /// x,y and z coordinates for each node
long *CX, *CY;    /// x,y and z code numbers for each node
long *CORN;
long  *XN,*XP,*YN,*YP,*ZN,*ZP;

long Nn;               /// number of nodes, number of elements
long Ndofn=2;          /// number of degrees of freedom in the node
int Param;             /// parameter for prescribed unit displacement in prescribed direction (x=1,y=2,z=3)
char *Fname;

/*
  Param: 0 ... pure periodic boundary conditions for prescribed unit stresses or strains
         1 ... symmetric boundary conditions with periodic b.c.'s on edges - perpendicular direction is fixed 
         2 ... constrained boundary conditions on edges
*/

int main(int argc,char *argv[])
{
  FILE *in,*out;

  long ii,jj,kk,n,nrn,nrn1,nrn2;
  long crn;
  long nxn,nxp,nyn,nyp;
  long codenum;
  double xn,xp,yn,yp;

  fprintf(stdout, "\n\n\n---  BOUNDARY NODES SELECTION OF 3D RECTANGULAR REPRESENTATIVE VOLUME ELEMENT  ---\n\n\n");
  if (argc < 4){
    fprintf (stderr,"\n Wrong number of command line parameters.");
    fprintf (stderr,"\n Use : %s input_file_name output_file_name prescribed_displacement_direction\n\n", argv[0]);
    return(1);
  }
  in = fopen(argv[1], "r");
  if (in==NULL){
    fprintf (stderr,"\n Input file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(2);
  }

  out = fopen(argv[2], "wt");
  if (out==NULL){
    fprintf (stderr,"\n Output file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(3);
  }
  Fname = argv[2];

  if (sscanf(argv[3], "%d", &Param) != 1)
  {
    fprintf (stderr,"\n Parameter for prescribed unit displacement has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(4);
  }

  //nodes reading
  //number of nodes:
  fscanf(in, "%ld",&Nn);

  //coordinates
  X = new double [Nn];
  Y = new double [Nn];
  Z = new double [Nn];

  XN = new long [Nn];
  XP = new long [Nn];

  YN = new long [Nn];
  YP = new long [Nn];

  for (ii = 0; ii < Nn; ii++){
    X[ii] = 0.0;
    Y[ii] = 0.0;
    Z[ii] = 0.0;
  }

  for (ii = 0; ii < Nn; ii++)
    {
      fscanf(in, "%ld",&n);
      fscanf(in, "%lf",&X[ii]);
      fscanf(in, "%lf",&Y[ii]);
      fscanf(in, "%lf",&Z[ii]);

      //round for 5 decimals
      X[ii] = X[ii]*100000.0;
      Y[ii] = Y[ii]*100000.0;
      Z[ii] = Z[ii]*100000.0;
      
      X[ii] = trunc(X[ii]);
      Y[ii] = trunc(Y[ii]);
      Z[ii] = trunc(Z[ii]);

      //X[ii] = floor(X[ii]);//debug??
      //Y[ii] = floor(Y[ii]);//debug??
      //Z[ii] = floor(Z[ii]);//debug??
      
      X[ii] = X[ii]/100000.0;
      Y[ii] = Y[ii]/100000.0;
      Z[ii] = Z[ii]/100000.0;
    }

  //sizes computation
  //minimal coordination selection:
  xn = X[0];
  yn = Y[0];
  for (ii = 0; ii < Nn; ii++)
    {
      if(X[ii] < xn)
	xn = X[ii];
      if(Y[ii] < yn)
	yn = Y[ii];
    }

  //maximal coordination selection:
  xp = X[0];
  yp = Y[0];
  for (ii = 0; ii < Nn; ii++)
    {
      if(X[ii] > xp)
	xp = X[ii];
      if(Y[ii] > yp)
	yp = Y[ii];
    }
  
  //output file printing
  //basic information about geometry reading
  fprintf(out, "\n # RVE dimensions:\n");
  fprintf(out, "# negative plane x-coordinate\n");
  fprintf(out, "%e\n",xn);//negative plane x-coordinate
  fprintf(out, "# positive plane x-coordinate\n");
  fprintf(out, "%e\n",xp);//posivite plane x-coordinate
  fprintf(out, "# negative plane y-coordinate\n");
  fprintf(out, "%e\n",yn);//negative plane y-coordinate
  fprintf(out, "# positive plane y-coordinate\n");
  fprintf(out, "%e\n",yp);//posivite plane y-coordinate

  //nodes printing
  fprintf(out, "\n # Number of nodes:\n");
  fprintf(out, "\n %ld \n",Nn);
  fprintf(out, "\n # Node number; x-coordinate, y-coordinate, z-coordinate:\n");
  for (ii = 0; ii < Nn; ii++)
    {
      fprintf(out, "  %ld",ii+1);
      fprintf(out, "  %e",X[ii]);
      fprintf(out, "  %e",Y[ii]);
      fprintf(out, "  %e\n",Z[ii]);
    }

  //corner nodes selection:
  CORN = new long [4];
  crn = 0;
  for (ii = 0; ii < Nn; ii++)
    {
      if((X[ii] == xn) && (Y[ii] == yn)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xp) && (Y[ii] == yn)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xn) && (Y[ii] == yp)){
	CORN[crn] = ii+1;
	crn++;
      }      
      if((X[ii] == xp) && (Y[ii] == yp)){
	CORN[crn] = ii+1;
	crn++;
      }
    }

  //corner nodes printing:
  fprintf(out, "\n# Number of corner nodes:\n");
  fprintf(out, "%ld \n",crn);
  fprintf(out, "#List of corner nodes:\n");
  for (ii = 0; ii < crn; ii++){
    fprintf(out, "%ld\n",CORN[ii]);
  }

  //edges selection:
  nxn = 0;
  nxp = 0;

  nyn = 0;
  nyp = 0;

  for (ii = 0; ii < Nn; ii++)
    {
      // x edges
      //XN:
      if(X[ii] == xn){
	XN[nxn] = ii+1;
	nxn++;
      }

      //XP:
      if(X[ii] == xp){
	XP[nxp] = ii+1;
	nxp++;
      }

      // y edges
      //YN:
      if(Y[ii] == yn){
	YN[nyn] = ii+1;
	nyn++;
      }

      //YP:
      if(Y[ii] == yp){
	YP[nyp] = ii+1;
	nyp++;
      }
    }
  
  //edges printing:
  fprintf(out, "\n# Number of x negative edge nodes:\n");
  fprintf(out, "\n %ld \n",nxn);
  fprintf(out, "\n# List of x negative edge nodes:\n");
  for (ii = 0; ii < nxn; ii++){
    fprintf(out, "%ld\n",XN[ii]);
  }

  fprintf(out, "\n# Number of x positive edge nodes:\n");
  fprintf(out, "\n %ld \n",nxp);
  fprintf(out, "\n# List of x positive edge nodes:\n");
  for (ii = 0; ii < nxp; ii++){
    fprintf(out, "%ld\n",XP[ii]);
  }

  fprintf(out, "\n# Number of y negative edge nodes:\n");
  fprintf(out, "\n %ld \n",nyn);
  fprintf(out, "\n# List of y negative edge nodes:\n");
  for (ii = 0; ii < nyn; ii++){
    fprintf(out, "%ld\n",YN[ii]);
  }

  fprintf(out, "\n# Number of y positive edge nodes:\n");
  fprintf(out, "\n %ld \n",nyp);
  fprintf(out, "\n# List of y positive edge nodes:\n");
  for (ii = 0; ii < nyp; ii++){
    fprintf(out, "%ld\n",YP[ii]);
  }

  //code numbers
  CX = new long [Nn];
  CY = new long [Nn];

  for (ii = 0; ii < Nn; ii++){
    CX[ii] = 1;
    CY[ii] = 1;
  }

  //reading and setup of corner nodes:
  for (ii = 0; ii < crn; ii++)
    {
      CX[CORN[ii]-1] = 0;
      CY[CORN[ii]-1] = 0;
    }

  if (Param < 2){
    codenum = 2;

    //setup of nodes on -x plane:
    for (jj = 0; jj < nxn; jj++){
      if(CX[XN[jj]-1] == 1){
	CX[XN[jj]-1] = codenum;
	codenum++;
      }
      if(CY[XN[jj]-1] == 1){
	CY[XN[jj]-1] = codenum;
	codenum++;
      }
    }
    
    //setup of nodes on +x plane:
    for (jj = 0; jj < nxp; jj++){
      for(kk = 0; kk < nxn; kk++)
	if((Y[XP[jj]-1] == Y[XN[kk]-1])){
	  if(CX[XP[jj]-1] == 1)
	    CX[XP[jj]-1] = CX[XN[kk]-1];
	  if(CY[XP[jj]-1] == 1)
	    CY[XP[jj]-1] = CY[XN[kk]-1];
	  break;
	}
    }
    
    //setup of nodes on -y plane:
    for (jj = 0; jj < nyn; jj++){
      if(CX[YN[jj]-1] == 1){
	CX[YN[jj]-1] = codenum;
	codenum++;
      }
      if(CY[YN[jj]-1] == 1){
	CY[YN[jj]-1] = codenum;
	codenum++;
      }
    }
    
    //setup of nodes on +y plane:
    for (jj = 0; jj < nyp; jj++){
      for(kk = 0; kk < nyn; kk++)
	if((X[YP[jj]-1] == X[YN[kk]-1])){
	  if(CX[YP[jj]-1] == 1)
	    CX[YP[jj]-1] = CX[YN[kk]-1];
	  if(CY[YP[jj]-1] == 1)
	    CY[YP[jj]-1] = CY[YN[kk]-1];
	  break;
	}
    }
  }

  //symmetric b.c:
  if ((Param == 1) || (Param == 2)){
    for (jj = 0; jj < Nn; jj++){
      if (X[jj] == xn)
	CX[jj] = 0;
      if (X[jj] == xp)
	CX[jj] = 0;
      if (Y[jj] == yn)
	CY[jj] = 0;
      if (Y[jj] == yp)
	CY[jj] = 0;
    }
  }
  
  //code numbers printing:
  fprintf(out,"\n# Code numbers \n");
  fprintf(out,"%ld\n",Nn);
  for(ii=0;ii<Nn;ii++)
    fprintf(out,"%ld  %ld  %ld\n",ii+1,CX[ii],CY[ii]);

  fclose(in);
  fclose(out);

  fprintf(stdout, "\n Boundary nodes have been printed into outputfile:  %s\n\n\n", Fname);

  return 0;
}
