#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

double *X, *Y, *Z;     /// x,y and z coordinates for each node
long *CX, *CY, *CZ;    /// x,y and z code numbers for each node
long *CORN;
long  *X1,*X2,*X3,*X4,*Y1,*Y2,*Y3,*Y4,*Z1,*Z2,*Z3,*Z4,*XN,*XP,*YN,*YP,*ZN,*ZP;

long Nn;               /// number of nodes, number of elements
long Ndofn=3;          /// number of degrees of freedom in the node
int Param;             /// parameter for prescribed unit displacement in prescribed direction (x=1,y=2,z=3)
char *Fname,*Fname2;

/*
  Param: 0 ... pure periodic boundary conditions for prescribed unit stresses or strains
         1 ... symmetric boundary conditions with periodic b.c.'s in planes - perpendicular direction is fixed 
         2 ... symmetric boundary conditions with free b.c.'s in planes - perpendicular direction is fixed 
         3 ... constraint boundary conditions
*/

int main(int argc,char *argv[])
{
  FILE *in,*out,*out2;

  long ii,jj,kk,n,nrn,nrn1,nrn2;
  long crn,nx1,nx2,nx3,nx4,ny1,ny2,ny3,ny4,nz1,nz2,nz3,nz4;
  long nxn,nxp,nyn,nyp,nzn,nzp;
  long codenum;
  double xn,xp,yn,yp,zn,zp;

  fprintf(stdout, "\n\n\n---  BOUNDARY NODES SELECTION OF 3D RECTANGULAR REPRESENTATIVE VOLUME ELEMENT  ---\n\n\n");
  if (argc < 5){
    fprintf (stderr,"\n Wrong number of command line parameters.");
    fprintf (stderr,"\n Use : %s input_file_name output_file_name.log output_file_name.con prescribed_displacement_direction\n\n", argv[0]);
    return(1);
  }
  in = fopen(argv[1], "r");
  if (in==NULL){
    fprintf (stderr,"\n Input file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(2);
  }

  out = fopen(argv[2], "wt");
  if (out==NULL){
    fprintf (stderr,"\n Output file .log has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(3);
  }
  Fname = argv[2];

  out2 = fopen(argv[3], "wt");
  if (out==NULL){
    fprintf (stderr,"\n Output file .con has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(3);
  }

  Fname2 = argv[3];

  if (sscanf(argv[4], "%d", &Param) != 1)
  {
    fprintf (stderr,"\n Parameter for prescribed unit displacement has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(4);
  }

  //nodes reading
  //number of nodes:
  fscanf(in, "%ld",&Nn);

  //coordinates
  X = new double [Nn];
  Y = new double [Nn];
  Z = new double [Nn];
  
  X1 = new long [Nn];
  X2 = new long [Nn];
  X3 = new long [Nn];
  X4 = new long [Nn];

  Y1 = new long [Nn];
  Y2 = new long [Nn];
  Y3 = new long [Nn];
  Y4 = new long [Nn];

  Z1 = new long [Nn];
  Z2 = new long [Nn];
  Z3 = new long [Nn];
  Z4 = new long [Nn];

  XN = new long [Nn];
  XP = new long [Nn];

  YN = new long [Nn];
  YP = new long [Nn];

  ZN = new long [Nn];
  ZP = new long [Nn];

  for (ii = 0; ii < Nn; ii++){
    X[ii] = 0.0;
    Y[ii] = 0.0;
    Z[ii] = 0.0;
  }

  for (ii = 0; ii < Nn; ii++)
    {
      fscanf(in, "%ld",&n);
      fscanf(in, "%lf",&X[ii]);
      fscanf(in, "%lf",&Y[ii]);
      fscanf(in, "%lf",&Z[ii]);

      //round for 5 decimals
      X[ii] = X[ii]*100000.0;
      Y[ii] = Y[ii]*100000.0;
      Z[ii] = Z[ii]*100000.0;
      
      X[ii] = trunc(X[ii]);
      Y[ii] = trunc(Y[ii]);
      Z[ii] = trunc(Z[ii]);

      //X[ii] = floor(X[ii]);//debug??
      //Y[ii] = floor(Y[ii]);//debug??
      //Z[ii] = floor(Z[ii]);//debug??
      
      X[ii] = X[ii]/100000.0;
      Y[ii] = Y[ii]/100000.0;
      Z[ii] = Z[ii]/100000.0;
    }

  //sizes computation
  //minimal coordination selection:
  xn = X[0];
  yn = Y[0];
  zn = Z[0];
  for (ii = 0; ii < Nn; ii++)
    {
      if(X[ii] < xn)
	xn = X[ii];
      if(Y[ii] < yn)
	yn = Y[ii];
      if(Z[ii] < zn)
	zn = Z[ii];      
    }

  //maximal coordination selection:
  xp = X[0];
  yp = Y[0];
  zp = Z[0];
  for (ii = 0; ii < Nn; ii++)
    {
      if(X[ii] > xp)
	xp = X[ii];
      if(Y[ii] > yp)
	yp = Y[ii];
      if(Z[ii] > zp)
	zp = Z[ii];      
    }

  //output file printing
  //basic information about geometry reading
  fprintf(out, "\n # RVE dimensions:\n");
  fprintf(out, "# negative plane x-coordinate\n");
  fprintf(out, "%e\n",xn);//negative plane x-coordinate
  fprintf(out, "# positive plane x-coordinate\n");
  fprintf(out, "%e\n",xp);//posivite plane x-coordinate
  fprintf(out, "# negative plane y-coordinate\n");
  fprintf(out, "%e\n",yn);//negative plane y-coordinate
  fprintf(out, "# positive plane y-coordinate\n");
  fprintf(out, "%e\n",yp);//posivite plane y-coordinate
  fprintf(out, "# negative plane z-coordinate\n");
  fprintf(out, "%e\n",zn);//negative plane z-coordinate
  fprintf(out, "# positive plane z-coordinate\n");
  fprintf(out, "%e\n",zp);//posivite plane z-coordinate

  //nodes printing
  fprintf(out2, "\n # Number of nodes:\n");
  fprintf(out2, "\n %ld \n",Nn);
  fprintf(out2, "\n # Node number; x-coordinate, y-coordinate, z-coordinate:\n");
  for (ii = 0; ii < Nn; ii++)
    {
      fprintf(out2, "  %ld",ii+1);
      fprintf(out2, "  %e",X[ii]);
      fprintf(out2, "  %e",Y[ii]);
      fprintf(out2, "  %e 3 0 0\n",Z[ii]);
    }

  //corner nodes selection:
  CORN = new long [8];
  crn = 0;
  for (ii = 0; ii < Nn; ii++)
    {
      //lower plane
      if((X[ii] == xn) && (Y[ii] == yn) && (Z[ii] == zn)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xp) && (Y[ii] == yn) && (Z[ii] == zn)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xn) && (Y[ii] == yp) && (Z[ii] == zn)){
	CORN[crn] = ii+1;
	crn++;
      }      
      if((X[ii] == xp) && (Y[ii] == yp) && (Z[ii] == zn)){
	CORN[crn] = ii+1;
	crn++;
      }

      //upper plane
      if((X[ii] == xn) && (Y[ii] == yn) && (Z[ii] == zp)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xp) && (Y[ii] == yn) && (Z[ii] == zp)){
	CORN[crn] = ii+1;
	crn++;
      }
      if((X[ii] == xn) && (Y[ii] == yp) && (Z[ii] == zp)){
	CORN[crn] = ii+1;
	crn++;
      }      
      if((X[ii] == xp) && (Y[ii] == yp) && (Z[ii] == zp)){
	CORN[crn] = ii+1;
	crn++;
      }
    }

  //corner nodes printing:
  fprintf(out, "\n# Number of corner nodes:\n");
  fprintf(out, "%ld \n",crn);
  fprintf(out, "#List of corner nodes:\n");
  for (ii = 0; ii < crn; ii++){
    fprintf(out, "%ld\n",CORN[ii]);
  }

  //edges selection:
  nx1 = 0;
  nx2 = 0;
  nx3 = 0;
  nx4 = 0;

  ny1 = 0;
  ny2 = 0;
  ny3 = 0;
  ny4 = 0;

  nz1 = 0;
  nz2 = 0;
  nz3 = 0;
  nz4 = 0;
  for (ii = 0; ii < Nn; ii++)
    {
      // x edges
      //X1:
      if((Y[ii] == yn) && (Z[ii] == zn)){
	X1[nx1] = ii+1;
	nx1++;
      }
      //X2:
      if((Y[ii] == yp) && (Z[ii] == zn)){
	X2[nx2] = ii+1;
	nx2++;
      }
      //X3:
      if((Y[ii] == yn) && (Z[ii] == zp)){
	X3[nx3] = ii+1;
	nx3++;
      }
      //X4:
      if((Y[ii] == yp) && (Z[ii] == zp)){
	X4[nx4] = ii+1;
	nx4++;
      }

      // y edges
      //Y1:
      if((X[ii] == xn) && (Z[ii] == zn)){
	Y1[ny1] = ii+1;
	ny1++;
      }
      //Y2:
      if((X[ii] == xp) && (Z[ii] == zn)){
	Y2[ny2] = ii+1;
	ny2++;
      }
      //Y3:
      if((X[ii] == xn) && (Z[ii] == zp)){
	Y3[ny3] = ii+1;
	ny3++;
      }
      //Y4:
      if((X[ii] == xp) && (Z[ii] == zp)){
	Y4[ny4] = ii+1;
	ny4++;
      }

      // z edges
      //Z1:
      if((X[ii] == xn) && (Y[ii] == yn)){
	Z1[nz1] = ii+1;
	nz1++;
      }
      //Z2:
      if((X[ii] == xp) && (Y[ii] == yn)){
	Z2[nz2] = ii+1;
	nz2++;
      }
      //Z3:
      if((X[ii] == xn) && (Y[ii] == yp)){
	Z3[nz3] = ii+1;
	nz3++;
      }
      //Z4:
      if((X[ii] == xp) && (Y[ii] == yp)){
	Z4[nz4] = ii+1;
	nz4++;
      }
    }

  //edges printing:
  fprintf(out, "\n# Number of x1 edge nodes:\n");
  fprintf(out, "\n %ld \n",nx1);
  fprintf(out, "#List of x1 nodes:\n");
  for (ii = 0; ii < nx1; ii++){
    fprintf(out, "%ld\n",X1[ii]);
    //fprintf(out, "%ld   %e\n",X1[ii],X[X1[ii]-1]);//debug??!!
  }

  fprintf(out, "\n# Number of x2 edge nodes:\n");
  fprintf(out, "\n %ld \n",nx2);
  fprintf(out, "#List of x2 nodes:\n");
  for (ii = 0; ii < nx2; ii++){
    fprintf(out, "%ld\n",X2[ii]);
    //fprintf(out, "%ld   %e\n",X2[ii],X[X2[ii]-1]);//debug??!!
  }

  fprintf(out, "\n# Number of x3 edge nodes:\n");
  fprintf(out, "\n %ld \n",nx3);
  fprintf(out, "#List of x3 nodes:\n");
  for (ii = 0; ii < nx3; ii++){
    fprintf(out, "%ld\n",X3[ii]);
    //fprintf(out, "%ld   %e\n",X3[ii],X[X3[ii]-1]);//debug??!!
  }

  fprintf(out, "\n# Number of x4 edge nodes:\n");
  fprintf(out, "\n %ld \n",nx4);
  fprintf(out, "#List of x4 nodes:\n");
  for (ii = 0; ii < nx4; ii++){
    fprintf(out, "%ld\n",X4[ii]);
    //fprintf(out, "%ld   %e\n",X4[ii],X[X4[ii]-1]);//debug??!!
  }

  fprintf(out, "\n# Number of y1 edge nodes:\n");
  fprintf(out, "\n %ld \n",ny1);
  fprintf(out, "#List of y1 nodes:\n");
  for (ii = 0; ii < ny1; ii++){
    fprintf(out, "%ld\n",Y1[ii]);
  }

  fprintf(out, "\n# Number of y2 edge nodes:\n");
  fprintf(out, "\n %ld \n",ny2);
  fprintf(out, "#List of y2 nodes:\n");
  for (ii = 0; ii < ny2; ii++){
    fprintf(out, "%ld\n",Y2[ii]);
  }

  fprintf(out, "\n# Number of y3 edge nodes:\n");
  fprintf(out, "\n %ld \n",ny3);
  fprintf(out, "#List of y3 nodes:\n");
  for (ii = 0; ii < ny3; ii++){
    fprintf(out, "%ld\n",Y3[ii]);
  }

  fprintf(out, "\n# Number of y4 edge nodes:\n");
  fprintf(out, "\n %ld \n",ny4);
  fprintf(out, "#List of y4 nodes:\n");
  for (ii = 0; ii < ny4; ii++){
    fprintf(out, "%ld\n",Y4[ii]);
  }


  fprintf(out, "\n# Number of z1 edge nodes:\n");
  fprintf(out, "\n %ld \n",nz1);
  fprintf(out, "#List of z1 nodes:\n");
  for (ii = 0; ii < nz1; ii++){
    fprintf(out, "%ld\n",Z1[ii]);
  }

  fprintf(out, "\n# Number of z2 edge nodes:\n");
  fprintf(out, "\n %ld \n",nz2);
  fprintf(out, "#List of z2 nodes:\n");
  for (ii = 0; ii < nz2; ii++){
    fprintf(out, "%ld\n",Z2[ii]);
  }

  fprintf(out, "\n# Number of z3 edge nodes:\n");
  fprintf(out, "\n %ld \n",nz3);
  fprintf(out, "#List of z3 nodes:\n");
  for (ii = 0; ii < nz3; ii++){
    fprintf(out, "%ld\n",Z3[ii]);
  }

  fprintf(out, "\n# Number of z4 edge nodes:\n");
  fprintf(out, "\n %ld \n",nz4);
  fprintf(out, "#List of z4 nodes:\n");
  for (ii = 0; ii < nz4; ii++){
    fprintf(out, "%ld\n",Z4[ii]);
  }


  //planes selection:
  nxn = 0;
  nxp = 0;

  nyn = 0;
  nyp = 0;

  nzn = 0;
  nzp = 0;
  for (ii = 0; ii < Nn; ii++)
    {
      // x planes
      //XN:
      if(X[ii] == xn){
	XN[nxn] = ii+1;
	nxn++;
      }

      //XP:
      if(X[ii] == xp){
	XP[nxp] = ii+1;
	nxp++;
      }

      // y planes
      //YN:
      if(Y[ii] == yn){
	YN[nyn] = ii+1;
	nyn++;
      }

      //YP:
      if(Y[ii] == yp){
	YP[nyp] = ii+1;
	nyp++;
      }

      // z planes
      //ZN:
      if(Z[ii] == zn){
	ZN[nzn] = ii+1;
	nzn++;
      }

      //ZP:
      if(Z[ii] == zp){
	ZP[nzp] = ii+1;
	nzp++;
      }
    }
 
  //planes printing:
  fprintf(out, "\n# Number of x negative plane nodes:\n");
  fprintf(out, "\n %ld \n",nxn);
  fprintf(out, "\n# List of x negative plane nodes:\n");
  for (ii = 0; ii < nxn; ii++){
    fprintf(out, "%ld\n",XN[ii]);
  }

  fprintf(out, "\n# Number of x positive plane nodes:\n");
  fprintf(out, "\n %ld \n",nxp);
  fprintf(out, "\n# List of x positive plane nodes:\n");
  for (ii = 0; ii < nxp; ii++){
    fprintf(out, "%ld\n",XP[ii]);
  }

  fprintf(out, "\n# Number of y negative plane nodes:\n");
  fprintf(out, "\n %ld \n",nyn);
  fprintf(out, "\n# List of y negative plane nodes:\n");
  for (ii = 0; ii < nyn; ii++){
    fprintf(out, "%ld\n",YN[ii]);
  }

  fprintf(out, "\n# Number of y positive plane nodes:\n");
  fprintf(out, "\n %ld \n",nyp);
  fprintf(out, "\n# List of y positive plane nodes:\n");
  for (ii = 0; ii < nyp; ii++){
    fprintf(out, "%ld\n",YP[ii]);
  }


  fprintf(out, "\n# Number of z negative plane nodes:\n");
  fprintf(out, "\n %ld \n",nzn);
  fprintf(out, "\n# List of z negative plane nodes:\n");
  for (ii = 0; ii < nzn; ii++){
    fprintf(out, "%ld\n",ZN[ii]);
  }

  fprintf(out, "\n# Number of z positive plane nodes:\n");
  fprintf(out, "\n %ld \n",nzp);
  fprintf(out, "\n# List of z positive plane nodes:\n");
  for (ii = 0; ii < nzp; ii++){
    fprintf(out, "%ld\n",ZP[ii]);
  }


  //code numbers
  CX = new long [Nn];
  CY = new long [Nn];
  CZ = new long [Nn];

  for (ii = 0; ii < Nn; ii++){
    CX[ii] = 1;
    CY[ii] = 1;
    CZ[ii] = 1;
  }

  //reading and setup of corner nodes:
  for (ii = 0; ii < crn; ii++)
    {
      CX[CORN[ii]-1] = 0;
      CY[CORN[ii]-1] = 0;
      CZ[CORN[ii]-1] = 0;
    }

  if (Param < 2){
    codenum = 2;
    //setup of nodes on x edges:
    //the first edge
    for (ii = 0; ii < nx1; ii++){
      if(CX[X1[ii]-1] == 1){
	CX[X1[ii]-1] = codenum;
	codenum++;
      }
      if(CY[X1[ii]-1] == 1){
	CY[X1[ii]-1] = codenum;
	codenum++;
      }
      if(CZ[X1[ii]-1] == 1){
	CZ[X1[ii]-1] = codenum;
	codenum++;
      }
    }
    
    //the second x edge
    for (jj = 0; jj < nx2; jj++){
      for(kk = 0; kk < nx1; kk++)
	if(X[X2[jj]-1] == X[X1[kk]-1]){
	  if(CX[X2[jj]-1] == 1)
	    CX[X2[jj]-1] = CX[X1[kk]-1];
	  if(CY[X2[jj]-1] == 1)
	    CY[X2[jj]-1] = CY[X1[kk]-1];
	  if(CZ[X2[jj]-1] == 1)
	    CZ[X2[jj]-1] = CZ[X1[kk]-1];
	  break;
	}
    }
    
    //the third x edge
    for (jj = 0; jj < nx3; jj++){
      for(kk = 0; kk < nx1; kk++)
	if(X[X3[jj]-1] == X[X1[kk]-1]){
	  if(CX[X3[jj]-1] == 1)
	    CX[X3[jj]-1] = CX[X1[kk]-1];
	  if(CY[X3[jj]-1] == 1)
	    CY[X3[jj]-1] = CY[X1[kk]-1];
	  if(CZ[X3[jj]-1] == 1)
	    CZ[X3[jj]-1] = CZ[X1[kk]-1];
	  break;
	}
    }
    
    //the fourth x edge
    for (jj = 0; jj < nx4; jj++){
      for(kk = 0; kk < nx1; kk++)
	if(X[X4[jj]-1] == X[X1[kk]-1]){
	  if(CX[X4[jj]-1] == 1)
	    CX[X4[jj]-1] = CX[X1[kk]-1];
	  if(CY[X4[jj]-1] == 1)
	    CY[X4[jj]-1] = CY[X1[kk]-1];
	  if(CZ[X4[jj]-1] == 1)
	    CZ[X4[jj]-1] = CZ[X1[kk]-1];
	  break;
	}
    }
    
    //setup of nodes on y edges:
    //the first edge
    for (ii = 0; ii < ny1; ii++){
      if(CX[Y1[ii]-1] == 1){
	CX[Y1[ii]-1] = codenum;
	codenum++;
      }
      if(CY[Y1[ii]-1] == 1){
	CY[Y1[ii]-1] = codenum;
	codenum++;
      }
      if(CZ[Y1[ii]-1] == 1){
	CZ[Y1[ii]-1] = codenum;
	codenum++;
      }
    }
    
    //the second y edge
    for (jj = 0; jj < ny2; jj++){
      for(kk = 0; kk < ny1; kk++)
	if(Y[Y2[jj]-1] == Y[Y1[kk]-1]){
	  if(CX[Y2[jj]-1] == 1)
	    CX[Y2[jj]-1] = CX[Y1[kk]-1];
	  if(CY[Y2[jj]-1] == 1)
	    CY[Y2[jj]-1] = CY[Y1[kk]-1];
	  if(CZ[Y2[jj]-1] == 1)
	    CZ[Y2[jj]-1] = CZ[Y1[kk]-1];
	  break;
	}
    }
    
    //the third y edge
    for (jj = 0; jj < ny3; jj++){
      for(kk = 0; kk < ny1; kk++)
	if(Y[Y3[jj]-1] == Y[Y1[kk]-1]){
	  if(CX[Y3[jj]-1] == 1)
	    CX[Y3[jj]-1] = CX[Y1[kk]-1];
	  if(CY[Y3[jj]-1] == 1)
	    CY[Y3[jj]-1] = CY[Y1[kk]-1];
	  if(CZ[Y3[jj]-1] == 1)
	    CZ[Y3[jj]-1] = CZ[Y1[kk]-1];
	  break;
	}
    }
    
    //the fourth y edge
    for (jj = 0; jj < ny4; jj++){
      for(kk = 0; kk < ny1; kk++)
	if(Y[Y4[jj]-1] == Y[Y1[kk]-1]){
	  if(CX[Y4[jj]-1] == 1)
	    CX[Y4[jj]-1] = CX[Y1[kk]-1];
	  if(CY[Y4[jj]-1] == 1)
	    CY[Y4[jj]-1] = CY[Y1[kk]-1];
	  if(CZ[Y4[jj]-1] == 1)
	    CZ[Y4[jj]-1] = CZ[Y1[kk]-1];
	  break;
	}
    }
    
    
    //setup of nodes on z edges:
    //the first edge
    for (ii = 0; ii < nz1; ii++){
      if(CX[Z1[ii]-1] == 1){
	CX[Z1[ii]-1] = codenum;
	codenum++;
      }
      if(CY[Z1[ii]-1] == 1){
	CY[Z1[ii]-1] = codenum;
	codenum++;
      }
      if(CZ[Z1[ii]-1] == 1){
	CZ[Z1[ii]-1] = codenum;
	codenum++;
      }
    }
    
    //the second z edge
    for (jj = 0; jj < nz2; jj++){
      for(kk = 0; kk < nz1; kk++)
	if(Z[Z2[jj]-1] == Z[Z1[kk]-1]){
	  if(CX[Z2[jj]-1] == 1)
	    CX[Z2[jj]-1] = CX[Z1[kk]-1];
	  if(CY[Z2[jj]-1] == 1)
	    CY[Z2[jj]-1] = CY[Z1[kk]-1];
	  if(CZ[Z2[jj]-1] == 1)
	    CZ[Z2[jj]-1] = CZ[Z1[kk]-1];
	  break;
	}
    }
    
    //the third z edge
    for (jj = 0; jj < nz3; jj++){
      for(kk = 0; kk < nz1; kk++)
	if(Z[Z3[jj]-1] == Z[Z1[kk]-1]){
	  if(CX[Z3[jj]-1] == 1)
	    CX[Z3[jj]-1] = CX[Z1[kk]-1];
	  if(CY[Z3[jj]-1] == 1)
	    CY[Z3[jj]-1] = CY[Z1[kk]-1];
	  if(CZ[Z3[jj]-1] == 1)
	    CZ[Z3[jj]-1] = CZ[Z1[kk]-1];
	  break;
	}
    }
    
    //the fourth z edge
    for (jj = 0; jj < nz4; jj++){
      for(kk = 0; kk < nz1; kk++)
	if(Z[Z4[jj]-1] == Z[Z1[kk]-1]){
	  if(CX[Z4[jj]-1] == 1)
	    CX[Z4[jj]-1] = CX[Z1[kk]-1];
	  if(CY[Z4[jj]-1] == 1)
	    CY[Z4[jj]-1] = CY[Z1[kk]-1];
	  if(CZ[Z4[jj]-1] == 1)
	    CZ[Z4[jj]-1] = CZ[Z1[kk]-1];
	  break;
	}
    }
    
    
    //setup of nodes on -x plane:
    for (jj = 0; jj < nxn; jj++){
      if(CX[XN[jj]-1] == 1){
	CX[XN[jj]-1] = codenum;
	codenum++;
      }
      if(CY[XN[jj]-1] == 1){
	CY[XN[jj]-1] = codenum;
	codenum++;
      }
      if(CZ[XN[jj]-1] == 1){
	CZ[XN[jj]-1] = codenum;
	codenum++;
      }
    }
    
    //setup of nodes on +x plane:
    for (jj = 0; jj < nxp; jj++){
      for(kk = 0; kk < nxn; kk++)
	if((Y[XP[jj]-1] == Y[XN[kk]-1]) && (Z[XP[jj]-1] == Z[XN[kk]-1])){
	  if(CX[XP[jj]-1] == 1)
	    CX[XP[jj]-1] = CX[XN[kk]-1];
	  if(CY[XP[jj]-1] == 1)
	    CY[XP[jj]-1] = CY[XN[kk]-1];
	  if(CZ[XP[jj]-1] == 1)
	    CZ[XP[jj]-1] = CZ[XN[kk]-1];
	  break;
	}
    }
    
    //setup of nodes on -y plane:
    for (jj = 0; jj < nyn; jj++){
      if(CX[YN[jj]-1] == 1){
	CX[YN[jj]-1] = codenum;
	codenum++;
      }
      if(CY[YN[jj]-1] == 1){
	CY[YN[jj]-1] = codenum;
	codenum++;
      }
      if(CZ[YN[jj]-1] == 1){
	CZ[YN[jj]-1] = codenum;
	codenum++;
      }
    }
    
    //setup of nodes on +y plane:
    for (jj = 0; jj < nyp; jj++){
      for(kk = 0; kk < nyn; kk++)
	if((X[YP[jj]-1] == X[YN[kk]-1]) && (Z[YP[jj]-1] == Z[YN[kk]-1])){
	  if(CX[YP[jj]-1] == 1)
	    CX[YP[jj]-1] = CX[YN[kk]-1];
	  if(CY[YP[jj]-1] == 1)
	    CY[YP[jj]-1] = CY[YN[kk]-1];
	  if(CZ[YP[jj]-1] == 1)
	    CZ[YP[jj]-1] = CZ[YN[kk]-1];
	  break;
	}
    }
    
    //setup of nodes on -z plane:
    for (jj = 0; jj < nzn; jj++){
      if(CX[ZN[jj]-1] == 1){
	CX[ZN[jj]-1] = codenum;
	codenum++;
      }
      if(CY[ZN[jj]-1] == 1){
	CY[ZN[jj]-1] = codenum;
	codenum++;
      }
      if(CZ[ZN[jj]-1] == 1){
	CZ[ZN[jj]-1] = codenum;
	codenum++;
      }
    }
    
    //reading and setup of nodes on +z plane:
    for (jj = 0; jj < nzp; jj++){
      for(kk = 0; kk < nzn; kk++)
	if((X[ZP[jj]-1] == X[ZN[kk]-1]) && (Y[ZP[jj]-1] == Y[ZN[kk]-1])){
	  if(CX[ZP[jj]-1] == 1)
	    CX[ZP[jj]-1] = CX[ZN[kk]-1];
	  if(CY[ZP[jj]-1] == 1)
	    CY[ZP[jj]-1] = CY[ZN[kk]-1];
	  if(CZ[ZP[jj]-1] == 1)
	    CZ[ZP[jj]-1] = CZ[ZN[kk]-1];
	  break;
	}
    }
  }

  //symmetric b.c:
  if ((Param == 1) || (Param == 2)){
    for (jj = 0; jj < Nn; jj++){
      if (X[jj] == xn)
	CX[jj] = 0;
      if (X[jj] == xp)
	CX[jj] = 0;
      if (Y[jj] == yn)
	CY[jj] = 0;
      if (Y[jj] == yp)
	CY[jj] = 0;
      if (Z[jj] == zn)
	CZ[jj] = 0;
      if (Z[jj] == zp)
	CZ[jj] = 0;  
    }
  }
   
  //symmetric b.c:
  if (Param == 3){
    for (jj = 0; jj < Nn; jj++){
      if ((X[jj] == xn) || (X[jj] == xp) || (Y[jj] == yn) || (Y[jj] == yp) || (Z[jj] == zn) || (Z[jj] == zp)){
	CX[jj] = 0;
	CX[jj] = 0;
	CY[jj] = 0;
	CY[jj] = 0;
	CZ[jj] = 0;
	CZ[jj] = 0;  
      }
    }
  }

  //code numbers printing:
  fprintf(out2,"\n# Code numbers \n");
  fprintf(out2,"%ld\n",Nn);
  for(ii=0;ii<Nn;ii++)
    fprintf(out2,"%ld  %ld  %ld  %ld\n",ii+1,CX[ii],CY[ii],CZ[ii]);

  //elements reading
  /*
    long Ne;
    
    fscanf(in,"%ld",&Ne);
    long *Node1, *Node2, *Node3, *Node4;
    Node1 = new long [Ne];
    Node2 = new long [Ne];
    Node3 = new long [Ne];
    Node4 = new long [Ne];
    double x0,x1,x2,x3,y0,y1,y2,y3,z0,z1,z2,z3,a0,a1,a2,b0,b1,b2,c0,c1,c2,D;
    long help;
    
    for(ii=0;ii<Ne;ii++)
    fscanf(in,"%ld %ld %ld  %ld  %ld  %ld\n",&help,&help,&Node1[ii],&Node2[ii],&Node3[ii],&Node4[ii]);
    
    //elements printing
    fprintf(out2, "\n # Number of elements:\n");
    fprintf(out2,"\n\n\n%ld\n",Ne);
    
    for(ii=0;ii<Ne;ii++){
    x0 = X[Node1[ii]];
    x1 = X[Node2[ii]];
    x2 = X[Node3[ii]];
    x3 = X[Node4[ii]];
    y0 = Y[Node1[ii]];
    y1 = Y[Node2[ii]];
    y2 = Y[Node3[ii]];
    y3 = Y[Node4[ii]];
    z0 = Z[Node1[ii]];
    z1 = Z[Node2[ii]];
    z2 = Z[Node3[ii]];
    z3 = Z[Node4[ii]];
    a0=x0-x3;
    a1=y0-y3;
    a2=z0-z3;
    b0=x1-x3;
    b1=y1-y3;
    b2=z1-z3;
    c0=x2-x3;
    c1=y2-y3;
    c2=z2-z3;
    D = (a0*b1*c2) + (b0*c1*a2) + (c0*a1*b2) - (c0*b1*a2) - (c1*b2*a0) - (c2*b0*a1);
    if(D > 0.0){
    fprintf(out2,"%ld  100 %ld  %ld  %ld  %ld 0    20          1     1 1    1 \n",ii+1,Node1[ii],Node2[ii],Node4[ii],Node3[ii]);
    printf("%ld %e\n",ii+1,D);
    }
    else
    fprintf(out2,"%ld  100 %ld  %ld  %ld  %ld 0    20          1     1 1    1 \n",ii+1,Node1[ii],Node2[ii],Node3[ii],Node4[ii]);      
  }
  */

  fclose(in);
  fclose(out);
  fclose(out2);

  fprintf(stdout, "\n Boundary nodes have been printed into outputfile:  %s\n\n\n", Fname2);
  fprintf(stdout, "\n Details about edges and planes have been printed into outputfile:  %s\n\n\n", Fname);

  return 0;
}
